# This file contains functions for defining the network formats
# Created by: Leonid Chindelevitch
# Last modified: October 27, 2014

from OutputProcessing import CreateSMatrix
from ModelProcessing import *
from Utilities import *
from fractions import Fraction
zero, one = Fraction(0), Fraction(1)

class Network:
    def __init__(self, species, metabolites, reactions, description = None):
        self.species = species
        self.metabolites = metabolites
        self.reactions = reactions
        self.compartments = sorted([x.compartment for x in self.metabolites])
        self.Matrix = None
        self.fullMatrix = None
        self.reducedMatrix = None
        self.reactionSubsets = None
        self.biomassCoefficients = [zero] * len(self.reactions)
        self.description = description
    def createMatrices(self):
        selfReacts = [reaction.pairs for reaction in self.reactions]
        numReacts = len(self.reactions)
        numMetabs = len(self.metabolites)
        self.fullMatrix = CreateSMatrix(selfReacts, numMetabs)
        intMetabs = [x for x in range(numMetabs) if not self.metabolites[x].external]
        self.Matrix = [self.fullMatrix[x] for x in intMetabs]
    def reduceNetwork(self, filename = 'Reduction.txt'):
        # Wrapper for reduceMatrix class
        N = self.Matrix
        Irr = [i for i, r in enumerate(self.reactions) if not r.reversible]
        # special = []
        reduction = reduceMatrix(N, Irr, filename)
        self.applyReduction(reduction)
    def applyReduction(self, reducedMatrix):
        self.reducedMatrix = reducedMatrix[0]
        irreversibles = reducedMatrix[1]
        metabStatuses = reducedMatrix[2]
        internalMetabs = [m for m in self.metabolites if not m.external]
        for i, metab in enumerate(internalMetabs):
            metab.reductionStatus = metabStatuses[i]
        reactStatuses = reducedMatrix[3]
        forwardOnly  = reducedMatrix[4]
        backwardOnly = reducedMatrix[5]
        subsets = reducedMatrix[6]
        self.reactionSubsets = []
        anchor_to_subset = {}
        for i, subset in enumerate(subsets):
            anchor = subset[0][0]
            anchor_to_subset[anchor] = subset
        anchors = anchor_to_subset.keys()
        index = 0
        for i, reaction in enumerate(self.reactions):
            reaction.reductionStatus = reactStatuses[i]
            if reactStatuses[i] == 0 or reactStatuses[i] == 4: # a reaction that survived (possibly a unidirectional one)
                geneCombo = self.reactions[reaction.index].geneCombination
                mult = one
                if reactStatuses[i] == 4:
                    if i in backwardOnly:
                        mult = -mult
                    elif i not in forwardOnly:
                        print('Error: this should never happen!')
                new_subset = Reaction_Subset([[reaction.index, mult]], index, not (index in irreversibles), geneCombo)
                index += 1
                self.reactionSubsets.append(new_subset)
            elif reactStatuses[i] == 5: # a true reaction subset (containing multiple reactions)
                if i in anchors:
                    subset = anchor_to_subset[i]
                    geneCombo = reduce(lambda u, v: u.logicalAnd(v), filter(None, [self.reactions[x].geneCombination for x in [y[0] for y in subset]]), CNF([]))
                    transformedSubset = [[y[0], y[1] * (1 - 2 * int(y[0] in backwardOnly))] for y in subset]
                    self.reactionSubsets.append(Reaction_Subset(transformedSubset, index, not (index in irreversibles), geneCombo))
                    index += 1
    def restoreZeroLoops(self, shortName): # adds zero loops at the end; uses a file derived from shortName to restore the subsets
        if self.reactionSubsets is None:
            print('Error: the network has not been reduced yet!')
            return
        zeroLoops = [i for i,r in enumerate(self.reactions) if r.reductionStatus == 7]
        curMatrix = transpose(self.reducedMatrix)
        zeros = [zero] * len(curMatrix[0])
        newMatrix = curMatrix + [zeros] * len(zeroLoops)
        self.reducedMatrix = transpose(newMatrix)
        if len(zeroLoops) > 0:
            for z in zeroLoops:
                self.reactions[z].reductionStatus = 5
            auxFile = shortName + 'ReductionFull.txt'
            f = open(auxFile)
            G = f.readlines()
            f.close()
            G = [x.strip().split('\t') for x in G]
            if G[-2] != ['']:
                Loops = [eval(x) for x in G[-2]]
                if not sorted(Loops) == sorted(zeroLoops):
                    print('Error: mismatch between file and reduction!')
                Subsets = []
                if G[-4] != ['']:
                    Subsets = [eval(x) for x in G[-4]]
                zeroSubsets = [x for x in Subsets if x[0][0] in Loops]
                index = len(self.reactionSubsets)
                backwardOnly  = []
                if G[-6] != ['']:
                    backwardOnly = [eval(x) for x in G[-6]]
                irreversibles = []
                if G[-14] != ['']:
                    irreversibles = [eval(x) for x in G[-14]]
                for subset in zeroSubsets:
                    geneCombo = reduce(lambda u, v: u.logicalAnd(v), filter(None, [self.reactions[x].geneCombination for x in [y[0] for y in subset]]), CNF([]))
                    transformedSubset = [[y[0], y[1] * (1 - 2 * int(y[0] in backwardOnly))] for y in subset]
                    self.reactionSubsets.append(Reaction_Subset(transformedSubset, index, not (index in irreversibles), geneCombo))
                    index += 1
    def findIrreversibleReactions(self):
        Irr = [i for i, r in enumerate(self.reactions) if not r.reversible]
        return Irr
    def findExchangeReactions(self):
        exchange = []
        externalInds = [i for i, m in enumerate(self.metabolites) if m.external]
        for ind, react in enumerate(self.reactions):
            for pair in react.pairs:
                if pair[0] in externalInds:
                    exchange.append(ind)
                    break
        return exchange
    def findExchangeReactionsReduced(self):
        exchange = self.findExchangeReactions()
        exchangeR = []
        if self.reactionSubsets is None:
            print('Error: the network has not been reduced yet!')
            return
        for ind, subset in enumerate(self.reactionSubsets):
            for pair in subset.pairs:
                if pair[0] in exchange:
                    exchangeR.append(ind)
                    break
        return exchangeR
    def findBiomassReaction(self):
        # This function tries to return the index of the biomass reaction
        if sum(self.biomassCoefficients) == one and one in self.biomassCoefficients:
            return self.biomassCoefficients.index(one)
        else:
            return -1
    def changeObjectiveFunction(self, coefficients):
        assert len(coefficients) == len(self.reactions)
        assert min(coefficients) >= 0
        self.biomassCoefficients = coefficients
        if self.reducedMatrix is not None:
            # convert the coefficient vector to match the reduced matrix
            pass
    def deleteReactions(self, reactionIndices):
        self.reactions = [r for i, r in enumerate(self.reactions) if i not in reactionIndices]
        self.createMatrices()
        if self.reducedMatrix is not None:
            # self.reduceNetwork()
            pass
    def addReaction(self, reactionName, reactionPairs, reversible = False, description = {}, biomass = False):
        nextIndex = len(self.reactions)
        self.reactions.append(Reaction(reactionName, reactionPairs, nextIndex, reversible, description = description))
        if biomass:
            self.biomassCoefficients.append(one)
        self.createMatrices()
        if self.reducedMatrix is not None:
            # self.reduceNetwork()
            pass
    def deleteMetabolites(self, metaboliteIndices):
        self.metabolites = [m for i, m in enumerate(self.metabolites) if i not in metaboliteIndices]
        self.createMatrices()
        if self.reducedMatrix is not None:
            # self.reduceNetwork()
            pass
    def addMetabolites(self, metaboliteName, metaboliteRepresentation):
        # need to worry about the species as well!
        pass
    def printReactionFormula(self, reactionNumber):
        curReact = self.reactions[reactionNumber]
        curPairs = curReact.pairs
        curMetabs = self.metabolites
        curSpecies = [x.species.name for x in curMetabs]
        curComparts = [x.compartment for x in curMetabs]
        curRev = curReact.reversible
        reagents = [x for x in curPairs if x[1] < 0]
        products = [x for x in curPairs if x[1] > 0]
        LHS = ' + '.join([str(-x[1]) + ' ' + curSpecies[x[0]] + '[' + curComparts[x[0]] + ']' for x in reagents])
        RHS = ' + '.join([str( x[1]) + ' ' + curSpecies[x[0]] + '[' + curComparts[x[0]] + ']' for x in products])
        if curRev:
            sep = ' <-> '
        else:
            sep = ' -> '
        return (LHS + sep + RHS)
    def adjustReversibility(self):
        if all([reaction.reversible for reaction in self.reactions]):
            print('All the reactions are currently reversible; trying to adjust using upper and lower bounds')
            descriptors = self.reactions[0].description.keys()
            UB = findExactlyOne(descriptors, 'upper')
            LB = findExactlyOne(descriptors, 'lower')
            if UB is not None and LB is not None:
                UB, LB = descriptors[UB], descriptors[LB]
                for ind, reaction in enumerate(self.reactions):
                    if ind % 100 == 0:
                        print('Processed ' + str(ind) + ' reactions so far')
                    curDescription = reaction.description
                    reaction.reversible = bool(float(curDescription[UB]) > 0 and float(curDescription[LB]) < 0)
            else:
                print('Error: cannot find either upper bounds or lower bounds in the reaction descriptions')
    def adjustReactionNames(self, oldName = 'EC numbers'):
        print('Trying to adjust reaction names')
        descriptors = self.reactions[0].description.keys()
        chosen = findExactlyOne(descriptors, 'name')
        if chosen is not None:
            goodDescriptor = descriptors[chosen]
            for ind, reaction in enumerate(self.reactions):
                if ind % 100 == 0:
                    print('Processed ' + str(ind) + ' reactions so far')
                if reaction.description is not None:
                    if reaction.name is not None:
                        reaction.description[oldName] = reaction.name
                    reaction.name = reaction.description[goodDescriptor]
                    del reaction.description[goodDescriptor]
    def adjustExternal(self, compart):
        print('Trying to adjust external metabolites')
        for metabolite in self.metabolites:
            if metabolite.compartment == compart:
                metabolite.external = True
        self.createMatrices()
    def adjustCompartments(self, compart = 'E'):
        if len(self.metabolites) == len(self.species):
            print('Trying to adjust compartments')
            speciesNames = [x.name for x in self.species]
            Map = {}
            for name in speciesNames:
                if name[1] == compart:
                    reducedName = name[:1] + name[2:]
                    Map[speciesNames.index(name)] = speciesNames.index(reducedName)
            for (key, value) in Map.iteritems():
                self.metabolites[key].species = self.species[value]
                self.metabolites[key].compartment = compart
                self.metabolites[key].external = True
            self.species = [self.species[x] for x in range(len(self.species)) if x not in Map]
        
class NetworkDescription():
    def __init__(self, genus, species, iden = None, authors = None, pmid = None, date = None):
        self.genus = genus
        self.species = species
        self.iden = iden
        self.authors = authors
        self.pmid = pmid
        self.date = date
    def search(self, s):
        #return a score corresponding to how close the match is
        matchVal = 0
        for t in s.split():
            t = t.lower()
            if t is self.genus.lower():
                matchVal += 15
            if t is self.species.lower():
                matchVal += 15
            if t == str(self.iden).lower():
                matchVal += 100
            if t is self.authors[0].lower():
                matchVal += 20
            if t is self.authors[-1].lower():
                matchVal += 10
            if t in [a.lower() for a in self.authors[1:-1]]:
                matchVal += 5
            if t[-2:] == self.date[-2:]:
                matchVal += 5
            if t == self.pmid:
                matchVal += 100
        return matchVal
    def write(self):
        print self.genus,self.species
        print self.iden
        print self.authors
        print self.pmid
        print self.date

class ExchangeStatus:
    IMPORT = 0
    EXPORT = 1
    BOTH = 2
    NONE = 3

class Reaction:
    def __init__(self, name, pairs, index = None, reversible = True, geneCombination = None, reductionStatus = None, exchangeStatus = None, description = None):
        self.name = name
        # pairs = [(metabolite_index,coefficient)]
        # omitted metabolites have coefficient 0
        self.pairs = pairs
        self.index = index
        self.reversible = reversible
        self.geneCombination = geneCombination
        self.reductionStatus = reductionStatus
        self.exchangeStatus = exchangeStatus
        self.description = description
    def show(self):
        print self.name, self.pairs, self.index, self.reversible, self.geneCombination, self.reductionStatus, self.exchangeStatus, self.description

class Metabolite:
    def __init__(self, species, compartment = None, index = None, external = False, reductionStatus = None, description = None):
        self.species = species
        self.compartment = compartment
        self.index = index
        self.external = external
        self.reductionStatus = reductionStatus
        self.description = description

class Species:
    def __init__(self, name = None, index = None, description = None):
        self.name = name
        self.index = index
        self.description = description

class Reaction_Subset:
    def __init__(self, pairs, index = None, reversible = True, geneCombination = None):
        self.pairs = pairs
        self.index = index
        self.reversible = reversible
        self.geneCombination = geneCombination

class CNF:
    def __init__(self, clauses):
        self.clauses = clauses
    def logicalAnd(self, other):
        new = CNF(clauses = self.clauses)
        new.clauses = new.clauses + other.clauses
        return new
    def logicalOr(self,other):
        new = CNF(clauses = self.clauses)
        full = DtoC(new.clauses) + DtoC(new.clauses)
        new.clauses = DtoC(full)
        new.logicalSimplify()
        return new
    def logicalSimplify(self):
        curClauses = self.clauses
        curClauses = [list(set(x)) for x in curClauses]
        curClauses = extractUnique(curClauses)
        self.clauses = curClauses
